////////////////////////////////////////////////////////////////////////////////
// ATL/AUX for ATL 7
// Public Domain
//
// Copyright (c) 1997-2005 by Andrew Nosenko <andien@geocities.com>,
// Portions copyright:
// (c) 1997-1998 by Computer Multimedia System, Inc.,
// (c) 1998-2001 by Mead & Co Limited
// (c) 2002-2006 by Emdat Inc.
//
// http://www.nozillium.com/atlaux/
// Version: 2.0.7
//

/* ~nab: 9/1/2010 3:16:51 pm
 * I only used parts used for hooking thunks, sadly they are for 32 bit programming only
 *
 * It is quite smart as it merely adds 'this' to ECX and leaves all other stack arguments
 * intact. I can't use ATL's CStdCallThunk (used for window creation) as it is intended 
 * for stdcall (static) callbacks and it moreover swaps the first stack argument for this
 * (see StartWindowProc etc). This isn't convenient for CBT hooks as the first argument 
 * is valuable and variable.
 *
 * I saw this guy doing a specific CBT hook replacement + storing of the lparam 3rd arg
 * http://www.codeproject.com/KB/COM/MouseProcThunk.aspx?msg=996483
 * But sadly this isn't for x64 either (although it is more readily expandable to x64 if
 * one finds the necessary assembly code. By contrast the ECX approach doesn't work in 
 * x64 as all calls there are fastcall (stack/register based, no special treatment of 
 * 'this' pointer
 * A similar approach is here http://www.codeproject.com/KB/cpp/thunk32.aspx
 *
 * _WIN64==SOLVED! @ 11/11/2013 11:27:23
 * using a less efficient method from stack overflow, see below
 */

#pragma once


#ifndef _WIN64

////////////////////////////////////////////////////////////////////////////////
/* Win32 Callback thunking, the main idea was taken from ATL.

Thunks could be used to turn an object+method pair into a closure that could be
passed to Win32 API requiring a callback address. This is a clever idea of ATL
team, I've only tried to propagate it over a general case.
Thunks are very suitable and efficient way to avoid tiresome TLS/CritSecs/maps
(when providing apartment threading and reentrancy).

CAuxThunk<> is provided for __thiscall methods, CAuxStdThunk<> is for __stdcall.

Here is an example of how to set up Win32 hook the reentrant way, without a
use of global variable:

class CHook:
  public CAuxThunk<CHook>
{
public:
  CHook(): m_hhook(NULL)
  {
    InitThunk((TMFP)CBTHook, this);
  }

  LRESULT CBTHook(int nCode, WPARAM wParam, LPARAM lParam);
  BOOL Hook() {
    m_hook = SetWindowsHookEx(WH_CBT, (HOOKPROC)GetThunk(), NULL, GetCurrentThreadId());
    return (BOOL)m_hook;
  }

  HHOOK m_hhook;
...
};

LRESULT CHook::CBTHook(int nCode, WPARAM wParam, LPARAM lParam)
{
  if ( nCode == HCBT_CREATEWND ) {
    UnhookWindowsHookEx(m_hook);
    HWND hwnd = (HWND)wParam;
    // do whatever we want with HWND
    ...
  }
  return CallNextHookEx(m_hook, nCode, wParam, lParam);
}

Another example. I need to process an event in my IE4 simple object
(windowless) *asynchronously* and use SetTimer API for that.
Unfortunately, TIMERPROC doesn't take any context info.
Here is how it has been solved:

struct TIMEOUT: CAuxThunk<TIMEOUT> {
  UINT m_timerID;
  CONTEXT m_contex;

  TIMEOUT(CONTEXT& contex): m_contex(contex)
  {
    InitThunk((TMFP)TimerProc, this);
  }
  void TimerProc(HWND, UINT, UINT idEvent, DWORD dwTime);
  ...
};

void TIMEOUT::TimerProc(HWND, UINT, UINT idEvent, DWORD dwTime)
{
  AuxKillTimer(NULL, m_timerID); // one-shot callback
  // do any processing task
  ...
  delete this;
}
HRESULT CSimpleObj::Post(CONTEXT& contex) {
  TIMEOUT* pTimeout = new TIMEOUT(context);
  pTimeout->m_timerID = ::SetTimer(NULL, 0, timeout, (TIMERPROC)pTimeout->GetThunk());
}
//////////////////////////////////////////////////////////////////////////////*/

#ifndef _M_IX86
  #pragma message("CAuxThunk/CAuxStdThunk is implemented for X86 only!")
#endif

#pragma pack(push, 1)

template <class T>
class CAuxThunk
{
  BYTE    m_mov;          // mov ecx, %pThis
  DWORD   m_this;         //
  BYTE    m_jmp;          // jmp func
  DWORD   m_relproc;      // relative jmp
public:
  typedef void (T::*TMFP)();

  void InitThunk(TMFP method, T* pThis)
  {
    union { DWORD func; TMFP method; } addr;
    addr.method = (TMFP)method;
    m_mov = 0xB9;
    m_this = reinterpret_cast<DWORD>(pThis);
    m_jmp = 0xE9;
    m_relproc = addr.func - reinterpret_cast<DWORD>(this+1);
		DWORD dwOldProt = 0;
		VirtualProtect(this, sizeof(*this), PAGE_EXECUTE_READWRITE, &dwOldProt);
    FlushInstructionCache(GetCurrentProcess(), this, sizeof(*this));
  }

  template <typename Method>
  void __stdcall InitThunk(Method method)
  {
		InitThunk((TMFP)method, static_cast<T*>(this));
	}

  FARPROC GetThunk() const {
    _ASSERTE(m_mov == 0xB9);
    return (FARPROC)this; }
};

template <class T>
class CAuxStdThunk
{
  BYTE    m_mov;          // mov eax, %pThis
  DWORD   m_this;         //
  DWORD   m_xchg_push;    // xchg eax, [esp] : push eax
  BYTE    m_jmp;          // jmp func
  DWORD   m_relproc;      // relative jmp
public:
  typedef void (__stdcall T::*TMFP)();
  void InitThunk(TMFP method, const T* pThis)
  {
    union {
      DWORD func;
      TMFP method;
    } addr;
    addr.method = method;
    m_mov = 0xB8;
    m_this = reinterpret_cast<DWORD>(pThis);
    m_xchg_push = 0x50240487;
    m_jmp = 0xE9;
    m_relproc = addr.func - reinterpret_cast<DWORD>(this+1);
		DWORD dwOldProt = 0;
		VirtualProtect(this, sizeof(*this), PAGE_EXECUTE_READWRITE, &dwOldProt);
    FlushInstructionCache(GetCurrentProcess(), this, sizeof(*this));
  }
  FARPROC GetThunk() const {
    _ASSERTE(m_mov == 0xB8);
    return (FARPROC)this; }
};

#pragma pack(pop) // CAuxThunk

#else // x64

// http://stackoverflow.com/questions/12136309/how-to-thunk-a-function-in-x86-and-x64-like-stdbind-in-c-but-dynamic
/* ARGUMENTS
 * param_count  number of arguments in the function f()
 *    @@@ looks like arguments must all be of pointer type (size) but that's ok for win64
 * buffer       will get filled with thunk code
 * n            number of fixed arguments (consecutive)
 * bound[]      respective fixed values
 * i            location of bound[] arg array within f() list
 * Fixed arguments are optional, if not required set i=n=bound=0 
 */
static size_t vbind(
    void *(/* cdecl, stdcall, or thiscall */ *f)(), size_t param_count,
    unsigned char buffer[/* >= 128 + n * (5 + sizeof(int) + sizeof(void*)) */],
    size_t const i, void *const bound[], unsigned int const n)
{
	unsigned char *p = buffer;
	unsigned char s = sizeof(void *);
	// ~nab: 2^log2(x) = x; b==3 for x64
	unsigned char const b = sizeof(int) == sizeof(void *) ? 2 : 3;  // log2(sizeof(void *))
	ATLASSERT(sizeof(int)==4); // ~nab: MSVC compiler
	ATLASSERT(i>= 0 && i+n <= param_count); // ~nab: bound args correctly placed?
	ATLASSERT(0==i || n); // it doesn't seem to matter but hey
	ATLASSERT(0==n || bound);
	BOOL x64 = b > 2;
	
	*p++ = 0x55;                                                                       // push     rbp
	if (x64) { *p++ = 0x48; } *p++ = 0x8B; *p++ = 0xEC;                                // mov      rbp, rsp
	if (x64)
	{
		*p++ = 0x48; *p++ = 0x89; *p++ = 0x4C; *p++ = 0x24; *p++ = 2 * s;               // mov      [rsp + 2 * s], rcx
		*p++ = 0x48; *p++ = 0x89; *p++ = 0x54; *p++ = 0x24; *p++ = 3 * s;               // mov      [rsp + 3 * s], rdx
		*p++ = 0x4C; *p++ = 0x89; *p++ = 0x44; *p++ = 0x24; *p++ = 4 * s;               // mov      [rsp + 4 * s], r8
		*p++ = 0x4C; *p++ = 0x89; *p++ = 0x4C; *p++ = 0x24; *p++ = 5 * s;               // mov      [rsp + 5 * s], r9
	}
	if (x64) { *p++ = 0x48; } *p++ = 0xBA; *(*(size_t **)&p)++ = param_count;          // mov      rdx, <param_count>
	if (x64) { *p++ = 0x48; } *p++ = 0x8B; *p++ = 0xC2;                                // mov      rax, rdx
	if (x64) { *p++ = 0x48; } *p++ = 0xC1; *p++ = 0xE0; *p++ = b;                      // shl      rax, log2(sizeof(void *))
	if (x64) { *p++ = 0x48; } *p++ = 0x2B; *p++ = 0xE0;                                // sub      rsp, rax
	*p++ = 0x57;                                                                       // push     rdi
	*p++ = 0x56;                                                                       // push     rsi
	*p++ = 0x51;                                                                       // push     rcx
	*p++ = 0x9C;                                                                       // pushfq
	if (x64) { *p++ = 0x48; } *p++ = 0xF7; *p++ = 0xD8;                                // neg      rax
	if (x64) { *p++ = 0x48; } *p++ = 0x8D; *p++ = 0x7C; *p++ = 0x05; *p++ = 0x00;      // lea      rdi, [rbp + rax]
	if (x64) { *p++ = 0x48; } *p++ = 0x8D; *p++ = 0x75; *p++ = 2 * s;                  // lea      rsi, [rbp + 10h]
	if (x64) { *p++ = 0x48; } *p++ = 0xB9; *(*(size_t **)&p)++ = i;                    // mov      rcx, <i>
	if (x64) { *p++ = 0x48; } *p++ = 0x2B; *p++ = 0xD1;                                // sub      rdx, rcx
	*p++ = 0xFC;                                                                       // cld
	*p++ = 0xF3; if (x64) { *p++ = 0x48; } *p++ = 0xA5;                                // rep movs [rdi], [rsi]
	for (unsigned int j = 0; j < n; j++)
	{
		unsigned int const o = j * sizeof(p);
		if (x64) { *p++ = 0x48; } *p++ = 0xB8; *(*(void ***)&p)++ = bound[j];           // mov      rax, <arg>
		if (x64) { *p++ = 0x48; } *p++ = 0x89; *p++ = 0x87; *(*(int **)&p)++ = o;       // mov      [rdi + <iArg>], rax
	}
	if (x64) { *p++ = 0x48; } *p++ = 0xB8; *(*(size_t **)&p)++ = n;                    // mov      rax, <count>
	if (x64) { *p++ = 0x48; } *p++ = 0x2B; *p++ = 0xD0;                                // sub      rdx, rax
	if (x64) { *p++ = 0x48; } *p++ = 0xC1; *p++ = 0xE0; *p++ = b;                      // shl      rax, log2(sizeof(void *))
	if (x64) { *p++ = 0x48; } *p++ = 0x03; *p++ = 0xF8;                                // add      rdi, rax
	if (x64) { *p++ = 0x48; } *p++ = 0x8B; *p++ = 0xCA;                                // mov      rcx, rdx
	*p++ = 0xF3; if (x64) { *p++ = 0x48; } *p++ = 0xA5;                                // rep movs [rdi], [rsi]
	*p++ = 0x9D;                                                                       // popfq
	*p++ = 0x59;                                                                       // pop      rcx
	*p++ = 0x5E;                                                                       // pop      rsi
	*p++ = 0x5F;                                                                       // pop      rdi
	if (x64)
	{
		*p++ = 0x48; *p++ = 0x8B; *p++ = 0x4C; *p++ = 0x24; *p++ = 0 * s;               // mov      rcx, [rsp + 0 * s]
		*p++ = 0x48; *p++ = 0x8B; *p++ = 0x54; *p++ = 0x24; *p++ = 1 * s;               // mov      rdx, [rsp + 1 * s]
		*p++ = 0x4C; *p++ = 0x8B; *p++ = 0x44; *p++ = 0x24; *p++ = 2 * s;               // mov      r8 , [rsp + 2 * s]
		*p++ = 0x4C; *p++ = 0x8B; *p++ = 0x4C; *p++ = 0x24; *p++ = 3 * s;               // mov      r9 , [rsp + 3 * s]
		*p++ = 0x48; *p++ = 0xB8; *(*(void *(***)())&p)++ = f;                          // mov      rax, <target_ptr>
		*p++ = 0xFF; *p++ = 0xD0;                                                       // call     rax
	}
	else
	{
		// if (thiscall) { *p++ = 0x59; }                                               // pop      rcx
		*p++ = 0xE8; *(*(ptrdiff_t **)&p)++ = (unsigned char *)f - (p + s);             // call     <fn_rel>
	}
	if (x64) { *p++ = 0x48; } *p++ = 0x8B; *p++ = 0xE5;                                // mov      rsp, rbp
	*p++ = 0x5D;                                                                       // pop      rbp
	*p++ = 0xC3;                                                                       // ret
	return p - &buffer[0];
}

#ifndef _WIN64
#error does not work for win32
#endif

// experimental x64 thiscall thunking
// maintain the same object interface as the original 32 bit class
template <class T>
class CAuxThunk
{
	// stub "code" is no longer the class itself
   BYTE m_thunk[1024]; // fixed; don't know size required apriori!
	// for 3 arguments the size required is under 200 bytes

public:
  typedef void (T::*TMFP)();

  // need an extra parameter, # of real arguments in the callback
  void InitThunk(TMFP method, T* pThis, int nArgs = 3)
  {
	  ATLASSERT(static_cast< CAuxThunk<T>* >(this) == pThis); // not really required as argument?
	  union { void* (*func)(); TMFP method; } addr;
	  addr.method = (TMFP)method;
	  ATLASSERT(sizeof(TMFP)==sizeof( void* (*)() )); // shouldn't be virtual
#if defined(_DEBUG) && defined(_WIN64)
	  // @@@ 64 bit compiler creates TMFP pointers that are 16 bytes long (!) even without virtuals
	  ATLASSERT(sizeof(TMFP)==sizeof(ULONGLONG)*2);
	  union {
		  struct {ULONGLONG a, b;};
		  TMFP method;
	  } ck_;
	  ck_.method = method;
	  ATLASSERT(0==ck_.b); // no useful information in high pointer?
#endif

	  // pass "this" as the first fixed argument
	  // that will put it on the stack instead of ECX, as should be in x64
	  void *fixed[] = { pThis };
	  size_t thunk_size = vbind(addr.func, nArgs+1, m_thunk, 0, fixed, 1);
	  ATLASSERT(thunk_size < sizeof(m_thunk));
	  
	  unsigned long old;
	  VirtualProtect(m_thunk, thunk_size, PAGE_EXECUTE_READWRITE, &old);
	  FlushInstructionCache(GetCurrentProcess(), m_thunk, thunk_size);
  }

  FARPROC GetThunk() const { return (FARPROC)(void*)m_thunk; }
};

#endif


//////////////////////////////////////////////////////////////////////////////*/
// sample class for WH_CBT callback thunking

#define CLASSNAME_LEN 100

// this can be used for all sorts of hooks, not just CBT! the callback prototype is identical
// for hooking OTHER processes also see SetWinEventHook
class CCBTHook: public CAuxThunk<CCBTHook>
{
public:
	CCBTHook(): m_hHook(NULL)
	{
#ifdef _DEBUG
		typedef LRESULT (CCBTHook::*HOOKSTUB)(int, WPARAM, LPARAM) ;
		HOOKSTUB t_ = &CCBTHook::CBTHook_stub;
		ATLASSERT(sizeof(t_)==sizeof(TMFP)); 
#endif
		InitThunk((TMFP)CBTHook_stub, this);
	}

	~CCBTHook() { 
		ATLASSERT(!m_hHook); // forgot to unhook?
		if(m_hHook)
			Hook(0);
	}

	BOOL Hook(BOOL on=TRUE) {
		if(on) {
			ATLASSERT(!m_hHook);
			// the hook is always attached on current thread, short term
			m_hHook = SetWindowsHookEx(WH_CBT, (HOOKPROC)GetThunk(), NULL, GetCurrentThreadId());
			return (BOOL)m_hHook;
		}
		else {
			ATLASSERT(m_hHook);
			if(m_hHook) {
				::UnhookWindowsHookEx(m_hHook);
				m_hHook = 0;
			}

			return 1;
		}
	}

	BOOL IsHooked() { return m_hHook != 0; }

	// helper to get class name (lowercase)
	LPCTSTR GetWndClass(WPARAM wParam, LPARAM lParam)
	{
		CBT_CREATEWND* pcw = (CBT_CREATEWND*)lParam;
		ATLASSERT(pcw);
		ATLASSERT(wParam);

		m_szClassName[0] = 0;
		LPCTSTR clname = pcw->lpcs->lpszClass;
		if(IS_INTRESOURCE(clname))
			::GetClassName((HWND)wParam, m_szClassName, CLASSNAME_LEN-1);
		else
			lstrcpy(m_szClassName, clname);

		ATLASSERT(m_szClassName[0]);
		CharLower(m_szClassName);
		return m_szClassName;
	}

	// override this for your processing
	virtual LRESULT HookProc(int nCode, WPARAM wParam, LPARAM lParam) 
	{
#ifdef _DEBUG // dump messages
		if(nCode == HCBT_CREATEWND) {
			LPCTSTR clname = GetWndClass(wParam, lParam);
			ATLTRACE(_T("hook createwnd %x class=%s\n"), wParam, clname);
		}
		else if(HCBT_SETFOCUS == nCode) {
			ATLTRACE(_T("hook setfocus %x lose=%x\n"), wParam, lParam);
		}
		else if(HCBT_ACTIVATE == nCode) {
			ATLTRACE(_T("hook activate %x\n"), wParam);
		}
		// none of these below are sent!
		else if(HCBT_MOVESIZE == nCode) {
			ATLTRACE(_T("hook movesize %x\n"), wParam);
		}
		else if(HCBT_MINMAX == nCode) {
			ATLTRACE(_T("hook minmax %x\n"), wParam);
		}
		else if(HCBT_QS == nCode) {
			ATLTRACE(_T("hook QS\n"));
		}
		else if(HCBT_SYSCOMMAND == nCode) {
			ATLTRACE(_T("hook syscom %x\n"), wParam);
		}
#endif

		return CallNextHookEx(m_hHook, nCode, wParam, lParam);
	}

protected:
  LRESULT CBTHook_stub(int nCode, WPARAM wParam, LPARAM lParam) {
	  ATLASSERT(m_hHook);
	  return HookProc(nCode, wParam, lParam); // virtual part
  }

  HHOOK m_hHook;
  TCHAR m_szClassName[CLASSNAME_LEN];
};
